######################################################################
#
# File: b2sdk/v1/download_dest.py
#
# Copyright 2019 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import io
import os
from abc import abstractmethod
from contextlib import contextmanager

from b2sdk._internal.stream.progress import WritingStreamWithProgress

from b2sdk._internal.utils import B2TraceMetaAbstract, limit_trace_arguments, set_file_mtime


class AbstractDownloadDestination(metaclass=B2TraceMetaAbstract):
    """
    Interface to a destination for a downloaded file.
    """

    @abstractmethod
    @limit_trace_arguments(
        skip=[
            'content_sha1',
        ]
    )
    def make_file_context(
        self,
        file_id,
        file_name,
        content_length,
        content_type,
        content_sha1,
        file_info,
        mod_time_millis,
        range_=None,
    ):
        """
        Return a context manager that yields a binary file-like object to use for
        writing the contents of the file.

        :param str file_id: the B2 file ID from the headers
        :param str file_name: the B2 file name from the headers
        :param str content_length: the content length
        :param str content_type: the content type from the headers
        :param str content_sha1: the content sha1 from the headers (or ``"none"`` for large files)
        :param dict file_info: the user file info from the headers
        :param int mod_time_millis: the desired file modification date in ms since 1970-01-01
        :param None,tuple[int,int] range_: starting and ending offsets of the received file contents.  Usually ``None``,
                       which means that the whole file is downloaded.
        :return: None
        """


class DownloadDestLocalFile(AbstractDownloadDestination):
    """
    Store a downloaded file into a local file and sets its modification time.
    """

    MODE = 'wb+'

    def __init__(self, local_file_path):
        self.local_file_path = local_file_path

    def make_file_context(
        self,
        file_id,
        file_name,
        content_length,
        content_type,
        content_sha1,
        file_info,
        mod_time_millis,
        range_=None,
    ):
        self.file_id = file_id
        self.file_name = file_name
        self.content_length = content_length
        self.content_type = content_type
        self.content_sha1 = content_sha1
        self.file_info = file_info
        self.range_ = range_
        return self.write_to_local_file_context(mod_time_millis)

    @contextmanager
    def write_to_local_file_context(self, mod_time_millis):
        completed = False
        try:
            # Open the file and let the caller write it.
            with open(self.local_file_path, self.MODE) as f:
                yield f

            set_file_mtime(self.local_file_path, mod_time_millis)

            # Set the flag that means to leave the downloaded file on disk.
            completed = True

        finally:
            # This is a best-effort attempt to clean up files that
            # failed to download, so we don't leave partial files
            # sitting on disk.
            if not completed:
                os.unlink(self.local_file_path)


class PreSeekedDownloadDest(DownloadDestLocalFile):
    """
    Store a downloaded file into a local file and sets its modification time.
    Does not truncate the target file, seeks to a given offset just after opening
    a descriptor.
    """

    MODE = 'rb+'

    def __init__(self, local_file_path, seek_target):
        self._seek_target = seek_target
        super().__init__(local_file_path)

    @contextmanager
    def write_to_local_file_context(self, *args, **kwargs):
        with super().write_to_local_file_context(*args, **kwargs) as f:
            f.seek(self._seek_target)
            yield f


class DownloadDestBytes(AbstractDownloadDestination):
    """
    Store a downloaded file into bytes in memory.
    """

    def __init__(self):
        self.bytes_written = None

    def make_file_context(
        self,
        file_id,
        file_name,
        content_length,
        content_type,
        content_sha1,
        file_info,
        mod_time_millis,
        range_=None,
    ):
        self.file_id = file_id
        self.file_name = file_name
        self.content_length = content_length
        self.content_type = content_type
        self.content_sha1 = content_sha1
        self.file_info = file_info
        self.mod_time_millis = mod_time_millis
        self.range_ = range_
        return self.capture_bytes_context()

    @contextmanager
    def capture_bytes_context(self):
        """
        Remember the bytes written in self.bytes_written.
        """
        # Make a place to store the data written
        bytes_io = io.BytesIO()

        # Let the caller write it
        yield bytes_io

        # Capture the result.  The BytesIO object won't let you grab
        # the data after it's closed
        self.bytes_written = bytes_io.getvalue()
        bytes_io.close()

    def get_bytes_written(self):
        if self.bytes_written is None:
            raise Exception('data not written yet')
        return self.bytes_written


class DownloadDestProgressWrapper(AbstractDownloadDestination):
    """
    Wrap a DownloadDestination and report progress to a ProgressListener.
    """

    def __init__(self, download_dest, progress_listener):
        self.download_dest = download_dest
        self.progress_listener = progress_listener

    def make_file_context(
        self,
        file_id,
        file_name,
        content_length,
        content_type,
        content_sha1,
        file_info,
        mod_time_millis,
        range_=None,
    ):
        return self.write_file_and_report_progress_context(
            file_id,
            file_name,
            content_length,
            content_type,
            content_sha1,
            file_info,
            mod_time_millis,
            range_,
        )

    @contextmanager
    def write_file_and_report_progress_context(
        self,
        file_id,
        file_name,
        content_length,
        content_type,
        content_sha1,
        file_info,
        mod_time_millis,
        range_,
    ):
        with self.download_dest.make_file_context(
            file_id,
            file_name,
            content_length,
            content_type,
            content_sha1,
            file_info,
            mod_time_millis,
            range_,
        ) as file_:
            total_bytes = content_length
            if range_ is not None:
                total_bytes = range_[1] - range_[0] + 1
            self.progress_listener.set_total_bytes(total_bytes)
            with self.progress_listener:
                yield WritingStreamWithProgress(file_, self.progress_listener)



######################################################################
#
# File: b2sdk/_internal/progress.py
#
# Copyright 2019 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import time
from abc import ABCMeta, abstractmethod

from .utils.escape import escape_control_chars

try:
    from tqdm import tqdm  # displays a nice progress bar
except ImportError:
    tqdm = None  # noqa


class AbstractProgressListener(metaclass=ABCMeta):
    """
    Interface expected by B2Api upload and download methods to report
    on progress.

    This interface just accepts the number of bytes transferred so far.
    Subclasses will need to know the total size if they want to report
    a percent done.
    """

    def __init__(self, description: str = ''):
        self.description = description
        self._closed = False

    @abstractmethod
    def set_total_bytes(self, total_byte_count: int) -> None:
        """
        Always called before __enter__ to set the expected total number of bytes.

        May be called more than once if an upload is retried.

        :param total_byte_count: expected total number of bytes
        """

    @abstractmethod
    def bytes_completed(self, byte_count: int) -> None:
        """
        Report the given number of bytes that have been transferred
        so far.  This is not a delta, it is the total number of bytes
        transferred so far.

        Transfer can fail and restart from the beginning, so byte count can
        decrease between calls.

        :param byte_count: number of bytes have been transferred
        """

    def _can_change_description(self) -> bool:
        """
        Determines, on a per-implementation basis, whether the description can be changed at this time.
        """
        return True

    def change_description(self, new_description: str) -> bool:
        """
        Ability to change the description after the listener is started.

        Note: whether the change of description is allowed depends on the implementation.
        The safest option is to change the description before setting the total bytes.

        :param new_description: the new description to be used
        :return: information whether the description was changed
        """
        if not self._can_change_description():
            return False

        self.description = new_description
        return True

    def close(self) -> None:
        """
        Must be called when you're done with the listener.
        In well-structured code, should be called only once.
        """
        # import traceback, sys; traceback.print_stack(file=sys.stdout)
        assert (
            self._closed is False
        ), 'progress listener was closed twice! uncomment the line above to debug this'
        self._closed = True

    def __enter__(self):
        """
        A standard context manager method.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        A standard context manager method.
        """
        self.close()


class TqdmProgressListener(AbstractProgressListener):
    """
    Progress listener based on tqdm library.

    This listener displays a nice progress bar, but requires `tqdm` package to be installed.
    """

    def __init__(self, *args, **kwargs):
        if tqdm is None:
            raise ModuleNotFoundError("No module named 'tqdm' found")
        self.tqdm = None  # set in set_total_bytes()
        self.prev_value = 0
        super().__init__(*args, **kwargs)

    def set_total_bytes(self, total_byte_count: int) -> None:
        if self.tqdm is None:
            self.tqdm = tqdm(
                desc=escape_control_chars(self.description),
                total=total_byte_count,
                unit='B',
                unit_scale=True,
                leave=True,
                miniters=1,
                smoothing=0.1,
                mininterval=0.2,
            )

    def bytes_completed(self, byte_count: int) -> None:
        # tqdm doesn't support running the progress bar backwards,
        # so on an upload retry, it just won't move until it gets
        # past the point where it failed.
        if self.prev_value < byte_count:
            self.tqdm.update(byte_count - self.prev_value)
            self.prev_value = byte_count

    def _can_change_description(self) -> bool:
        return self.tqdm is None

    def close(self) -> None:
        if self.tqdm is not None:
            self.tqdm.close()
        super().close()


class SimpleProgressListener(AbstractProgressListener):
    """
    Just a simple progress listener which prints info on a console.
    """

    def __init__(self, *args, **kwargs):
        self.complete = 0
        self.last_time = time.time()
        self.any_printed = False
        self.total = 0  # set in set_total_bytes()
        super().__init__(*args, **kwargs)

    def set_total_bytes(self, total_byte_count: int) -> None:
        self.total = total_byte_count

    def bytes_completed(self, byte_count: int) -> None:
        now = time.time()
        elapsed = now - self.last_time
        if 3 <= elapsed and self.total != 0:
            if not self.any_printed:
                print(escape_control_chars(self.description))
            print('     %d%%' % int(100.0 * byte_count / self.total))
            self.last_time = now
            self.any_printed = True

    def _can_change_description(self) -> bool:
        return not self.any_printed

    def close(self) -> None:
        if self.any_printed:
            print('    DONE.')
        super().close()


class DoNothingProgressListener(AbstractProgressListener):
    """
    This listener gives no output whatsoever.
    """

    def set_total_bytes(self, total_byte_count: int) -> None:
        pass

    def bytes_completed(self, byte_count: int) -> None:
        pass


class ProgressListenerForTest(AbstractProgressListener):
    """
    Capture all the calls so they can be checked.
    """

    def __init__(self, *args, **kwargs):
        self.calls = []
        super().__init__(*args, **kwargs)

    def set_total_bytes(self, total_byte_count: int) -> None:
        self.calls.append('set_total_bytes(%d)' % (total_byte_count,))

    def bytes_completed(self, byte_count: int) -> None:
        self.calls.append('bytes_completed(%d)' % (byte_count,))

    def close(self) -> None:
        self.calls.append('close()')
        super().close()

    def get_calls(self) -> list[str]:
        return self.calls


def make_progress_listener(description: str, quiet: bool) -> AbstractProgressListener:
    """
    Produce the best progress listener available for the given parameters.

    :param description: listener description
    :param quiet: if ``True``, do not output anything
    :return: a listener object
    """
    if quiet:
        return DoNothingProgressListener()
    elif tqdm is not None:
        return TqdmProgressListener(description)
    else:
        return SimpleProgressListener(description)



######################################################################
#
# File: b2sdk/v3/exception.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from b2sdk._internal.account_info.exception import AccountInfoError
from b2sdk._internal.account_info.exception import CorruptAccountInfo
from b2sdk._internal.account_info.exception import MissingAccountData
from b2sdk._internal.exception import (
    AccessDenied,
    AlreadyFailed,
    B2ConnectionError,
    B2Error,
    B2HttpCallbackException,
    B2HttpCallbackPostRequestException,
    B2HttpCallbackPreRequestException,
    B2RequestTimeout,
    B2RequestTimeoutDuringUpload,
    B2SimpleError,
    BadDateFormat,
    BadFileInfo,
    BadJson,
    BadRequest,
    BadUploadUrl,
    BrokenPipe,
    BucketIdNotFound,
    BucketNotAllowed,
    CapabilityNotAllowed,
    CapExceeded,
    ChecksumMismatch,
    ClockSkew,
    Conflict,
    ConnectionReset,
    CopyArgumentsMismatch,
    DestFileNewer,
    DestinationDirectoryDoesntAllowOperation,
    DestinationDirectoryDoesntExist,
    DestinationIsADirectory,
    DestinationParentIsNotADirectory,
    DisablingFileLockNotSupported,
    DuplicateBucketName,
    EmailNotVerified,
    FileAlreadyHidden,
    FileNameNotAllowed,
    FileNotPresent,
    FileSha1Mismatch,
    InvalidAuthToken,
    InvalidJsonResponse,
    InvalidMetadataDirective,
    InvalidRange,
    InvalidUploadSource,
    MaxFileSizeExceeded,
    MaxRetriesExceeded,
    MissingPart,
    NonExistentBucket,
    NoPaymentHistory,
    NotAllowedByAppKeyError,
    PartSha1Mismatch,
    PotentialS3EndpointPassedAsRealm,
    RestrictedBucket,
    RestrictedBucketMissing,
    RetentionWriteError,
    ServiceError,
    SourceReplicationConflict,
    SSECKeyError,
    SSECKeyIdMismatchInCopy,
    StorageCapExceeded,
    TooManyRequests,
    TransactionCapExceeded,
    TransientErrorMixin,
    TruncatedOutput,
    Unauthorized,
    UnexpectedCloudBehaviour,
    UnknownError,
    UnknownHost,
    UnrecognizedBucketType,
    UnsatisfiableRange,
    UnusableFileName,
    WrongEncryptionModeForBucketDefault,
    interpret_b2_error,
)
from b2sdk._internal.scan.exception import EmptyDirectory
from b2sdk._internal.scan.exception import EnvironmentEncodingError
from b2sdk._internal.scan.exception import InvalidArgument
from b2sdk._internal.scan.exception import NotADirectory
from b2sdk._internal.scan.exception import UnableToCreateDirectory
from b2sdk._internal.scan.exception import UnsupportedFilename
from b2sdk._internal.scan.exception import check_invalid_argument
from b2sdk._internal.sync.exception import IncompleteSync

__all__ = (
    'AccessDenied',
    'AccountInfoError',
    'AlreadyFailed',
    'B2ConnectionError',
    'B2Error',
    'B2HttpCallbackException',
    'B2HttpCallbackPostRequestException',
    'B2HttpCallbackPreRequestException',
    'B2RequestTimeout',
    'B2RequestTimeoutDuringUpload',
    'B2SimpleError',
    'BadDateFormat',
    'BadFileInfo',
    'BadJson',
    'BadRequest',
    'BadUploadUrl',
    'BrokenPipe',
    'BucketIdNotFound',
    'BucketNotAllowed',
    'CapabilityNotAllowed',
    'CapExceeded',
    'ChecksumMismatch',
    'ClockSkew',
    'Conflict',
    'ConnectionReset',
    'CopyArgumentsMismatch',
    'CorruptAccountInfo',
    'DestFileNewer',
    'DestinationDirectoryDoesntAllowOperation',
    'DestinationDirectoryDoesntExist',
    'DestinationIsADirectory',
    'DestinationParentIsNotADirectory',
    'DisablingFileLockNotSupported',
    'DuplicateBucketName',
    'EmailNotVerified',
    'EmptyDirectory',
    'EnvironmentEncodingError',
    'FileAlreadyHidden',
    'FileNameNotAllowed',
    'FileNotPresent',
    'FileSha1Mismatch',
    'IncompleteSync',
    'InvalidArgument',
    'InvalidAuthToken',
    'InvalidJsonResponse',
    'InvalidMetadataDirective',
    'InvalidRange',
    'InvalidUploadSource',
    'MaxFileSizeExceeded',
    'MaxRetriesExceeded',
    'MissingAccountData',
    'MissingPart',
    'NonExistentBucket',
    'NoPaymentHistory',
    'NotADirectory',
    'NotAllowedByAppKeyError',
    'PartSha1Mismatch',
    'PotentialS3EndpointPassedAsRealm',
    'RestrictedBucket',
    'RestrictedBucketMissing',
    'RetentionWriteError',
    'ServiceError',
    'SourceReplicationConflict',
    'StorageCapExceeded',
    'TooManyRequests',
    'TransactionCapExceeded',
    'TransientErrorMixin',
    'TruncatedOutput',
    'UnableToCreateDirectory',
    'Unauthorized',
    'UnexpectedCloudBehaviour',
    'UnknownError',
    'UnknownHost',
    'UnrecognizedBucketType',
    'UnsatisfiableRange',
    'UnsupportedFilename',
    'UnusableFileName',
    'interpret_b2_error',
    'check_invalid_argument',
    'SSECKeyIdMismatchInCopy',
    'SSECKeyError',
    'WrongEncryptionModeForBucketDefault',
)



######################################################################
#
# File: b2sdk/_internal/version_utils.py
#
# Copyright 2019 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import inspect
import re
import warnings
from abc import ABCMeta, abstractmethod
from functools import total_ordering, wraps

from b2sdk.version import VERSION


@total_ordering
class _Version:
    """
    Rudimentary semver version parser.

    It uses VERY naive parsing which is only supposed to produce a tuple, able to
    compare major.minor.patch versions.
    It does not support PEP 440 epoch, pre-releases, post-releases, local versions, etc.
    """

    def __init__(self, version: str):
        self._raw = version
        self._parsed = self._parse_version(version)

    def __str__(self):
        return self._raw

    def __eq__(self, other):
        return self._parsed == other._parsed

    def __lt__(self, other):
        return self._parsed < other._parsed

    @classmethod
    def _parse_version(cls, version: str) -> tuple[int, ...]:
        if '!' in version:  # strip PEP 440 epoch
            version = version.split('!', 1)[1]
        return tuple(map(int, re.findall(r'\d+', version)))


class AbstractVersionDecorator(metaclass=ABCMeta):
    WHAT = NotImplemented  # 'function', 'method', 'class' etc

    def __init__(self, changed_version, cutoff_version=None, reason='', current_version=None):
        """
        Changed_version, cutoff_version and current_version are version strings.
        """
        if current_version is None:  # this is for tests only
            current_version = VERSION  # TODO autodetect by going up the qualname tree and trying getattr(part, '__version__')
        self.current_version = _Version(current_version)  #: current version
        self.reason = reason

        self.changed_version = self._parse_if_not_none(
            changed_version
        )  #: version in which the decorator was added
        self.cutoff_version = self._parse_if_not_none(
            cutoff_version
        )  #: version in which the decorator (and something?) shall be removed

    @classmethod
    def _parse_if_not_none(cls, version):
        if version is None:
            return None
        return _Version(version)

    @abstractmethod
    def __call__(self, func):
        """
        The actual implementation of decorator. Needs self.source to be set before it's called.
        """
        if self.cutoff_version and self.changed_version:
            assert (
                self.changed_version < self.cutoff_version
            ), f'{self.__class__.__name__} decorator is set to start renaming {self.WHAT} {self.source!r} starting at version {self.changed_version} and finishing in {self.cutoff_version}. It needs to start at a lower version and finish at a higher version.'


class AbstractDeprecator(AbstractVersionDecorator):
    ALTERNATIVE_DECORATOR = NotImplemented

    def __init__(self, target, *args, **kwargs):
        self.target = target
        super().__init__(*args, **kwargs)


class rename_argument(AbstractDeprecator):
    """
    Change the argument name to new one if old one is used, warns about deprecation in docs and through a warning.

    >>> @rename_argument('aaa', 'bbb', '0.1.0', '0.2.0')
    >>> def easy(bbb):
    >>>     return bbb

    >>> easy(aaa=5)
    'aaa' is a deprecated argument for 'easy' function/method - it was renamed to 'bbb' in version 0.1.0. Support for the old name is going to be dropped in 0.2.0.
    5
    >>>
    """

    WHAT = 'argument'
    ALTERNATIVE_DECORATOR = 'discourage_argument'

    def __init__(self, source, *args, **kwargs):
        self.source = source
        super().__init__(*args, **kwargs)

    def __call__(self, func):
        super().__call__(func)
        signature = inspect.signature(func)
        has_target_arg = self.target in signature.parameters or any(
            p.kind == p.VAR_KEYWORD for p in signature.parameters.values()
        )
        assert has_target_arg, f'{self.target!r} is not an argument of the decorated function so it cannot be remapped to from a deprecated parameter name'

        @wraps(func)
        def wrapper(*args, **kwargs):
            if self.source in kwargs:
                assert (
                    self.target not in kwargs
                ), f'both argument names were provided: {self.source!r} (deprecated) and {self.target!r} (new)'
                kwargs[self.target] = kwargs[self.source]
                del kwargs[self.source]
                info = f'{self.source!r} is a deprecated argument for {func.__name__!r} function/method - it was renamed to {self.target!r}'
                if self.changed_version:
                    info += f' in version {self.changed_version}'
                if self.cutoff_version:
                    info += f'. Support for the old name is going to be dropped in {self.cutoff_version}'

                warnings.warn(
                    f'{info}.',
                    DeprecationWarning,
                )
            return func(*args, **kwargs)

        return wrapper


class rename_function(AbstractDeprecator):
    """
    Warn about deprecation in docs and through a DeprecationWarning when used.  Use it to decorate a proxy function, like this:

    >>> def new(foobar):
    >>>     return foobar ** 2
    >>> @rename_function(new, '0.1.0', '0.2.0')
    >>> def old(foo, bar):
    >>>     return new(foo + bar)
    >>> old()
    'old' is deprecated since version 0.1.0 - it was moved to 'new', please switch to use that. The proxy for the old name is going to be removed in 0.2.0.
    123
    >>>

    """

    WHAT = 'function'
    ALTERNATIVE_DECORATOR = 'discourage_function'

    def __init__(self, target, *args, **kwargs):
        if callable(target):
            target = target.__name__
        super().__init__(target, *args, **kwargs)

    def __call__(self, func):
        self.source = func.__name__
        super().__call__(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f'{func.__name__!r} is deprecated since version {self.changed_version} - it was moved to {self.target!r}, please switch to use that. The proxy for the old name is going to be removed in {self.cutoff_version}.',
                DeprecationWarning,
            )
            return func(*args, **kwargs)

        return wrapper


class rename_method(rename_function):
    WHAT = 'method'
    ALTERNATIVE_DECORATOR = 'discourage_method'


class FeaturePreviewWarning(FutureWarning):
    """
    Feature Preview Warning

    Marks a feature, that is in "Feature Preview" state.
    Such features are not yet fully stable and are subject to change or even outright removal.
    Do not rely on them in production code.
    """



######################################################################
#
# File: b2sdk/_internal/account_info/__init__.py
#
# Copyright 2019 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from .in_memory import InMemoryAccountInfo

assert InMemoryAccountInfo



######################################################################
#
# File: b2sdk/_internal/utils/filesystem.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
import pathlib
import platform
import stat

_IS_WINDOWS = platform.system() == 'Windows'


def points_to_fifo(path: pathlib.Path) -> bool:
    """Check if the path points to a fifo."""
    path = path.resolve()
    try:
        return stat.S_ISFIFO(path.stat().st_mode)
    except OSError:
        return False


_STDOUT_FILENAME = 'CON' if _IS_WINDOWS else '/dev/stdout'
STDOUT_FILEPATH = pathlib.Path(_STDOUT_FILENAME)


def points_to_stdout(path: pathlib.Path) -> bool:
    """Check if the path points to stdout."""
    try:
        return path == STDOUT_FILEPATH or path.resolve() == STDOUT_FILEPATH
    except OSError:
        return False



######################################################################
#
# File: b2sdk/v2/exception.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from b2sdk.v3.exception import *  # noqa

v3BucketIdNotFound = BucketIdNotFound
UnSyncableFilename = UnsupportedFilename


# overridden to retain old style isinstance check and attributes
class BucketIdNotFound(v3BucketIdNotFound, BadRequest):
    def __init__(self, bucket_id):
        super().__init__(bucket_id)
        self.message = f'Bucket with id={bucket_id} not found'
        self.code = 'bad_bucket_id'

    def __str__(self):
        return BadRequest.__str__(self)


class RestrictedBucket(B2Error):
    def __init__(self, bucket_name):
        super().__init__()
        self.bucket_name = bucket_name

    def __str__(self):
        return 'Application key is restricted to bucket: %s' % self.bucket_name



######################################################################
#
# File: b2sdk/_internal/transfer/outbound/large_file_upload_state.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import threading


class LargeFileUploadState:
    """
    Track the status of uploading a large file, accepting updates
    from the tasks that upload each of the parts.

    The aggregated progress is passed on to a ProgressListener that
    reports the progress for the file as a whole.

    This class is THREAD SAFE.
    """

    def __init__(self, file_progress_listener):
        """
        :param b2sdk.v2.AbstractProgressListener file_progress_listener: a progress listener object to use. Use :py:class:`b2sdk.v2.DoNothingProgressListener` to disable.
        """
        self.lock = threading.RLock()
        self.error_message = None
        self.file_progress_listener = file_progress_listener
        self.part_number_to_part_state = {}
        self.bytes_completed = 0

    def set_error(self, message):
        """
        Set an error message.

        :param str message: an error message
        """
        with self.lock:
            self.error_message = message

    def has_error(self):
        """
        Check whether an error occurred.

        :rtype: bool
        """
        with self.lock:
            return self.error_message is not None

    def get_error_message(self):
        """
        Fetche an error message.

        :return: an error message
        :rtype: str
        """
        with self.lock:
            assert self.has_error()
            return self.error_message

    def update_part_bytes(self, bytes_delta):
        """
        Update listener progress info.

        :param int bytes_delta: number of bytes to increase a progress for
        """
        with self.lock:
            self.bytes_completed += bytes_delta
            self.file_progress_listener.bytes_completed(self.bytes_completed)



######################################################################
#
# File: b2sdk/version.py
#
# Copyright 2019 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from importlib.metadata import version as _version
from sys import version_info as _version_info

__all__ = [
    'VERSION',
    'PYTHON_VERSION',
    'USER_AGENT',
]

VERSION = _version('b2sdk')

PYTHON_VERSION = '.'.join(map(str, _version_info[:3]))  # something like: 3.9.1

USER_AGENT = f'backblaze-b2/{VERSION} python/{PYTHON_VERSION}'



######################################################################
#
# File: b2sdk/_internal/testing/helpers/bucket_manager.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import logging
from collections.abc import Iterable
from datetime import datetime, timedelta
from itertools import chain
from typing import Any

import tenacity

from b2sdk._internal.api import B2Api
from b2sdk._internal.bucket import Bucket
from b2sdk._internal.exception import BadRequest, BucketIdNotFound, FileNotPresent, TooManyRequests
from b2sdk._internal.file_lock import NO_RETENTION_FILE_SETTING, LegalHold, RetentionMode
from b2sdk._internal.testing.helpers.buckets import (
    BUCKET_CREATED_AT_MILLIS,
    BUCKET_NAME_LENGTH,
    GENERAL_BUCKET_NAME_PREFIX,
    NODE_DESCRIPTION,
    bucket_name_part,
)
from b2sdk._internal.utils import current_time_millis

ONE_HOUR_MILLIS = 60 * 60 * 1000
BUCKET_CLEANUP_PERIOD_MILLIS = timedelta(hours=3).total_seconds() * 1000

logger = logging.getLogger(__name__)


class BucketManager:
    def __init__(
        self,
        dont_cleanup_old_buckets: bool,
        b2_api: B2Api,
        current_run_prefix: str = '',
        general_prefix: str = GENERAL_BUCKET_NAME_PREFIX,
    ):
        self.current_run_prefix = current_run_prefix
        self.general_prefix = general_prefix
        self.dont_cleanup_old_buckets = dont_cleanup_old_buckets
        self.b2_api = b2_api
        self.bucket_name_log: list[str] = []

    def new_bucket_name(self) -> str:
        bucket_name = self.current_run_prefix + bucket_name_part(
            BUCKET_NAME_LENGTH - len(self.current_run_prefix)
        )
        self.bucket_name_log.append(bucket_name)
        return bucket_name

    def new_bucket_info(self) -> dict:
        return {
            BUCKET_CREATED_AT_MILLIS: str(current_time_millis()),
            'created_by': NODE_DESCRIPTION,
        }

    def create_bucket(self, bucket_type: str = 'allPublic', **kwargs) -> Bucket:
        bucket_name = kwargs.pop('name', self.new_bucket_name())
        return self.b2_api.create_bucket(
            bucket_name,
            bucket_type=bucket_type,
            bucket_info=self.new_bucket_info(),
            **kwargs,
        )

    def _should_remove_bucket(self, bucket: Bucket) -> tuple[bool, str]:
        if self.current_run_prefix and bucket.name.startswith(self.current_run_prefix):
            return True, 'it is a bucket for this very run'
        if self.dont_cleanup_old_buckets:
            return False, 'old buckets ought not to be cleaned'
        if bucket.name.startswith(self.general_prefix):
            if BUCKET_CREATED_AT_MILLIS in bucket.bucket_info:
                delete_older_than = current_time_millis() - BUCKET_CLEANUP_PERIOD_MILLIS
                this_bucket_creation_time = int(bucket.bucket_info[BUCKET_CREATED_AT_MILLIS])
                if this_bucket_creation_time < delete_older_than:
                    return (
                        True,
                        f'this_bucket_creation_time={this_bucket_creation_time} < delete_older_than={delete_older_than}',
                    )
                return (
                    False,
                    f'this_bucket_creation_time={this_bucket_creation_time} >= delete_older_than={delete_older_than}',
                )
            else:
                return True, 'undefined ' + BUCKET_CREATED_AT_MILLIS
        return False, f'does not start with {self.general_prefix!r}'

    def clean_buckets(self, quick=False):
        # even with use_cache=True, if cache is empty API call will be made
        buckets = self.b2_api.list_buckets(use_cache=quick)
        remaining_buckets = []
        for bucket in buckets:
            should_remove, why = self._should_remove_bucket(bucket)
            if not should_remove:
                print(f'Skipping bucket removal {bucket.name!r} because {why}')
                remaining_buckets.append(bucket)
                continue

            print('Trying to remove bucket:', bucket.name, 'because', why)
            try:
                self.clean_bucket(bucket)
            except BucketIdNotFound:
                print(f'It seems that bucket {bucket.name} has already been removed')
        print('Total bucket count after cleanup:', len(remaining_buckets))
        for bucket in remaining_buckets:
            print(bucket)

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(TooManyRequests),
        wait=tenacity.wait_exponential(),
        stop=tenacity.stop_after_attempt(8),
    )
    def clean_bucket(
        self,
        bucket: Bucket | str,
        only_files: bool = False,
        only_folders: list[str] | None = None,
        ignore_retentions: bool = False,
    ):
        """
        Clean contents of bucket, by default also deleting the bucket.

        Args:
            bucket (Bucket | str): Bucket object or name
            only_files (bool): If to only delete files and not the bucket
            only_folders (list[str] | None): If not None, filter to only files in given folders.
            ignore_retentions (bool): If deletion should happen regardless of files' retention mode.
        """
        if isinstance(bucket, str):
            bucket = self.b2_api.get_bucket_by_name(bucket)

        if not only_files:
            # try optimistic bucket removal first, since it is completely free (as opposed to `ls` call)
            try:
                return self.b2_api.delete_bucket(bucket)
            except BucketIdNotFound:
                return  # bucket was already removed
            except BadRequest as exc:
                assert exc.code == 'cannot_delete_non_empty_bucket'

        files_leftover = False

        file_versions: Iterable[Any]
        if only_folders:
            file_versions = chain.from_iterable(
                [
                    bucket.ls(
                        path=folder,
                        latest_only=False,
                        recursive=True,
                    )
                    for folder in only_folders
                ]
            )
        else:
            file_versions = bucket.ls(latest_only=False, recursive=True)

        for file_version_info, _ in file_versions:
            if file_version_info.file_retention and not ignore_retentions:
                if file_version_info.file_retention.mode == RetentionMode.GOVERNANCE:
                    print('Removing retention from file version:', file_version_info.id_)
                    self.b2_api.update_file_retention(
                        file_version_info.id_,
                        file_version_info.file_name,
                        NO_RETENTION_FILE_SETTING,
                        True,
                    )
                elif file_version_info.file_retention.mode == RetentionMode.COMPLIANCE:
                    if file_version_info.file_retention.retain_until > current_time_millis():
                        print(
                            f'File version: {file_version_info.id_} cannot be removed due to compliance mode retention'
                        )
                        files_leftover = True
                        continue
                elif file_version_info.file_retention.mode == RetentionMode.NONE:
                    pass
                else:
                    raise ValueError(
                        f'Unknown retention mode: {file_version_info.file_retention.mode}'
                    )
            if file_version_info.legal_hold.is_on():
                print('Removing legal hold from file version:', file_version_info.id_)
                self.b2_api.update_file_legal_hold(
                    file_version_info.id_, file_version_info.file_name, LegalHold.OFF
                )
            print('Removing file version:', file_version_info.id_)
            try:
                self.b2_api.delete_file_version(file_version_info.id_, file_version_info.file_name)
            except FileNotPresent:
                print(
                    f'It seems that file version {file_version_info.id_} has already been removed'
                )

        if files_leftover:
            print('Unable to remove bucket because some retained files remain')
        elif not only_files:
            print('Removing bucket:', bucket.name)
            try:
                self.b2_api.delete_bucket(bucket)
            except BucketIdNotFound:
                print(f'It seems that bucket {bucket.name} has already been removed')
        print()

    def count_and_print_buckets(self) -> int:
        buckets = self.b2_api.list_buckets()
        count = len(buckets)
        print(f'Total bucket count at {datetime.now()}: {count}')
        for i, bucket in enumerate(buckets, start=1):
            print(f'- {i}\t{bucket.name} [{bucket.id_}]')
        return count



######################################################################
#
# File: b2sdk/_internal/testing/fixtures/api.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import http
import http.client
import os

import pytest

from b2sdk._internal.testing.helpers.api import authorize, get_b2_auth_data, get_realm


@pytest.fixture(scope='session')
def realm():
    yield get_realm()


@pytest.fixture(autouse=True, scope='session')
def set_http_debug():
    if os.environ.get('B2_DEBUG_HTTP'):
        http.client.HTTPConnection.debuglevel = 1


@pytest.fixture(scope='session')
def b2_auth_data():
    try:
        return get_b2_auth_data()
    except ValueError as ex:
        pytest.fail(ex.args[0])


@pytest.fixture(scope='session')
def _b2_api(b2_auth_data, realm):
    b2_api, _ = authorize(b2_auth_data, realm)
    return b2_api


@pytest.fixture(scope='session')
def b2_api(_b2_api, bucket_manager):
    return _b2_api



######################################################################
#
# File: b2sdk/v3/testing/__init__.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

# testing - it is not imported in v3.__init__ as it depends on pytest and other test dependencies.

from b2sdk._internal.testing.helpers.api import get_b2_auth_data, authorize, get_realm
from b2sdk._internal.testing.helpers.base import IntegrationTestBase
from b2sdk._internal.testing.helpers.buckets import (
    GENERAL_BUCKET_NAME_PREFIX,
    BUCKET_NAME_LENGTH,
    BUCKET_CREATED_AT_MILLIS,
    NODE_DESCRIPTION,
    RNG,
    RNG_COUNTER,
    RNG_SEED,
    get_bucket_name_prefix,
    random_token,
)
from b2sdk._internal.testing.helpers.bucket_manager import (
    ONE_HOUR_MILLIS,
    BUCKET_CLEANUP_PERIOD_MILLIS,
    BucketManager,
)
from b2sdk._internal.testing.fixtures.api import (
    set_http_debug,
    b2_auth_data,
    _b2_api,
    b2_api,
    realm,
)
from b2sdk._internal.testing.fixtures.buckets import (
    pytest_addoption,
    dont_cleanup_old_buckets,
    bucket_name_prefix,
    general_bucket_name_prefix,
    bucket_manager,
    bucket,
    b2_subfolder,
)



######################################################################
#
# File: b2sdk/_internal/testing/helpers/base.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import pytest

from b2sdk._internal.testing.helpers.bucket_manager import BucketManager
from b2sdk.v2 import B2Api
from b2sdk.v2.exception import DuplicateBucketName


@pytest.mark.usefixtures('cls_setup')
class IntegrationTestBase:
    b2_api: B2Api
    this_run_bucket_name_prefix: str
    bucket_manager: BucketManager

    @pytest.fixture(autouse=True, scope='class')
    def cls_setup(self, request, b2_api, b2_auth_data, bucket_name_prefix, bucket_manager):
        cls = request.cls
        cls.b2_auth_data = b2_auth_data
        cls.this_run_bucket_name_prefix = bucket_name_prefix
        cls.bucket_manager = bucket_manager
        cls.b2_api = b2_api
        cls.info = b2_api.account_info

    @pytest.fixture(autouse=True)
    def setup_method(self):
        self.buckets_created = []
        yield
        for bucket in self.buckets_created:
            self.bucket_manager.clean_bucket(bucket)

    def write_zeros(self, file, number):
        line = b'0' * 1000 + b'\n'
        line_len = len(line)
        written = 0
        while written <= number:
            file.write(line)
            written += line_len

    def create_bucket(self):
        bucket_name = self.bucket_manager.new_bucket_name()
        try:
            bucket = self.bucket_manager.create_bucket(name=bucket_name)
        except DuplicateBucketName:
            self._duplicated_bucket_name_debug_info(bucket_name)
            raise
        self.buckets_created.append(bucket)
        return bucket

    def _duplicated_bucket_name_debug_info(self, bucket_name: str) -> None:
        # Trying to obtain as much information as possible about this bucket.
        print(' DUPLICATED BUCKET DEBUG START '.center(60, '='))
        bucket = self.b2_api.get_bucket_by_name(bucket_name)

        print('Bucket metadata:')
        bucket_dict = bucket.as_dict()
        for info_key, info in bucket_dict.items():
            print(f'\t{info_key}: "{info}"')

        print('All files (and their versions) inside the bucket:')
        ls_generator = bucket.ls(recursive=True, latest_only=False)
        for file_version, _directory in ls_generator:
            # as_dict() is bound to have more info than we can use,
            # but maybe some of it will cast some light on the issue.
            print(f'\t{file_version.file_name} ({file_version.as_dict()})')

        print(' DUPLICATED BUCKET DEBUG END '.center(60, '='))



######################################################################
#
# File: b2sdk/_pyinstaller/hook-b2sdk.py
#
# Copyright 2021 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from PyInstaller.utils.hooks import copy_metadata

datas = copy_metadata('b2sdk')



######################################################################
#
# File: b2sdk/v2/version_utils.py
#
# Copyright 2023 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from b2sdk import v3


class _OldAbstractDeprecatorMixin:
    def __call__(self, *args, **kwargs):
        if self.cutoff_version:
            assert (
                self.current_version < self.cutoff_version
            ), f'{self.__class__.__name__} decorator is still used in version {self.current_version} when old {self.WHAT} name {self.source!r} was scheduled to be dropped in {self.cutoff_version}. It is time to remove the mapping.'
        ret = super().__call__(*args, **kwargs)
        assert (
            self.changed_version <= self.current_version
        ), f'{self.__class__.__name__} decorator indicates that the replacement of {self.WHAT} {self.source!r} should take place in the future version {self.changed_version}, while the current version is {self.cutoff_version}. It looks like should be _discouraged_ at this point and not _deprecated_ yet. Consider using {self.ALTERNATIVE_DECORATOR!r} decorator instead.'
        return ret


class rename_argument(_OldAbstractDeprecatorMixin, v3.rename_argument):
    pass


class rename_function(_OldAbstractDeprecatorMixin, v3.rename_function):
    pass



######################################################################
#
# File: b2sdk/v0/account_info.py
#
# Copyright 2019 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

from b2sdk._internal import version_utils
from b2sdk import v1


class OldAccountInfoMethods:
    """this class contains proxy methods for deprecated signatures renamed for consistency in mid-2019"""

    def get_account_id_or_app_key_id(self):
        """
        Return the application key ID used to authenticate.

        :rtype: str

        .. deprecated:: 0.1.6
           Use :func:`get_application_key_id` instead.
        """
        return self.get_application_key_id()

    @version_utils.rename_argument(
        'account_id_or_app_key_id',
        'application_key_id',
        '0.1.5',
        None,
    )
    def set_auth_data(
        self,
        account_id,
        auth_token,
        api_url,
        download_url,
        minimum_part_size,
        application_key,
        realm,
        allowed=None,
        application_key_id=None,
        s3_api_url=None,
    ):
        # we need to enumerate all the parameters and cannot just "*args, **kwargs" because
        # the deprecation decorator doesn't feel safe with the kwargs approach
        return super().set_auth_data(
            account_id,
            auth_token,
            api_url,
            download_url,
            minimum_part_size,
            application_key,
            realm,
            allowed,
            application_key_id,
            s3_api_url=s3_api_url,
        )


class AbstractAccountInfo(OldAccountInfoMethods, v1.AbstractAccountInfo):
    pass


class InMemoryAccountInfo(OldAccountInfoMethods, v1.InMemoryAccountInfo):
    pass


class UrlPoolAccountInfo(OldAccountInfoMethods, v1.UrlPoolAccountInfo):
    pass


class SqliteAccountInfo(OldAccountInfoMethods, v1.SqliteAccountInfo):
    pass



######################################################################
#
# File: b2sdk/_internal/utils/range_.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import dataclasses
import re

_RANGE_HEADER_RE = re.compile(
    r'^(?:bytes[ =])?(?P<start>\d+)-(?P<end>\d+)(?:/(?:(?P<complete_length>\d+)|\*))?$'
)


@dataclasses.dataclass(eq=True, order=True, frozen=True)
class Range:
    """
    HTTP ranges use an *inclusive* index at the end.
    """

    __slots__ = ['start', 'end']

    start: int
    end: int

    def __post_init__(self):
        assert 0 <= self.start <= self.end or (
            self.start == 1 and self.end == 0
        ), f'Invalid range: {self}'

    @classmethod
    def from_header(cls, raw_range_header: str) -> Range:
        """
        Factory method which returns an object constructed from Range http header.

        raw_range_header example: 'bytes=0-11'
        """
        return cls.from_header_with_size(raw_range_header)[0]

    @classmethod
    def from_header_with_size(cls, raw_range_header: str) -> tuple[Range, int | None]:
        """
        Factory method which returns an object constructed from Range http header.

        raw_range_header example: 'bytes=0-11'
        """
        match = _RANGE_HEADER_RE.match(raw_range_header)
        if not match:
            raise ValueError(f'Invalid range header: {raw_range_header}')
        start = int(match.group('start'))
        end = int(match.group('end'))
        complete_length = match.group('complete_length')
        complete_length = int(complete_length) if complete_length else None
        return cls(start, end), complete_length

    def size(self) -> int:
        return self.end - self.start + 1

    def subrange(self, sub_start, sub_end) -> Range:
        """
        Return a range that is part of this range.

        :param sub_start: index relative to the start of this range.
        :param sub_end: (Inclusive!) index relative to the start of this range.
        :return: a new Range
        """
        assert 0 <= sub_start <= sub_end < self.size()
        return self.__class__(self.start + sub_start, self.start + sub_end)

    def as_tuple(self) -> tuple[int, int]:
        return self.start, self.end

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.start}, {self.end})'


EMPTY_RANGE = Range(1, 0)



######################################################################
#
# File: b2sdk/_internal/transfer/outbound/copy_manager.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import logging

from b2sdk._internal.encryption.setting import EncryptionMode, EncryptionSetting
from b2sdk._internal.exception import AlreadyFailed, CopyArgumentsMismatch, SSECKeyIdMismatchInCopy
from b2sdk._internal.file_lock import FileRetentionSetting, LegalHold
from b2sdk._internal.http_constants import SSE_C_KEY_ID_FILE_INFO_KEY_NAME
from b2sdk._internal.progress import AbstractProgressListener
from b2sdk._internal.raw_api import MetadataDirectiveMode
from b2sdk._internal.transfer.transfer_manager import TransferManager
from b2sdk._internal.utils.thread_pool import ThreadPoolMixin

logger = logging.getLogger(__name__)


class CopyManager(TransferManager, ThreadPoolMixin):
    """
    Handle complex actions around server side copy to free raw_api from that responsibility.
    """

    MAX_LARGE_FILE_SIZE = 10 * 1000 * 1000 * 1000 * 1000  # 10 TB

    @property
    def account_info(self):
        return self.services.session.account_info

    def copy_file(
        self,
        copy_source,
        file_name,
        content_type,
        file_info,
        destination_bucket_id,
        progress_listener,
        destination_encryption: EncryptionSetting | None = None,
        source_encryption: EncryptionSetting | None = None,
        legal_hold: LegalHold | None = None,
        file_retention: FileRetentionSetting | None = None,
    ):
        # Run small copies in the same thread pool as large file copies,
        # so that they share resources during a sync.
        return self._thread_pool.submit(
            self._copy_small_file,
            copy_source,
            file_name,
            content_type=content_type,
            file_info=file_info,
            destination_bucket_id=destination_bucket_id,
            progress_listener=progress_listener,
            destination_encryption=destination_encryption,
            source_encryption=source_encryption,
            legal_hold=legal_hold,
            file_retention=file_retention,
        )

    def copy_part(
        self,
        large_file_id,
        part_copy_source,
        part_number,
        large_file_upload_state,
        finished_parts=None,
        destination_encryption: EncryptionSetting | None = None,
        source_encryption: EncryptionSetting | None = None,
    ):
        return self._thread_pool.submit(
            self._copy_part,
            large_file_id,
            part_copy_source,
            part_number,
            large_file_upload_state,
            finished_parts=finished_parts,
            destination_encryption=destination_encryption,
            source_encryption=source_encryption,
        )

    def _copy_part(
        self,
        large_file_id,
        part_copy_source,
        part_number,
        large_file_upload_state,
        finished_parts,
        destination_encryption: EncryptionSetting | None,
        source_encryption: EncryptionSetting | None,
    ):
        """
        Copy a file part to started large file.

        :param :param str bucket_id: a bucket ID
        :param large_file_id: a large file ID
        :param b2sdk.v2.CopySource part_copy_source: copy source that represents a range (not necessarily a whole file)
        :param b2sdk.v2.LargeFileUploadState large_file_upload_state: state object for progress reporting
                                                                      on large file upload
        :param dict,None finished_parts: dictionary of known finished parts, keys are part numbers,
                                         values are instances of :class:`~b2sdk.v2.Part`
        :param b2sdk.v2.EncryptionSetting destination_encryption: encryption settings for the destination
                        (``None`` if unknown)
        :param b2sdk.v2.EncryptionSetting source_encryption: encryption settings for the source
                        (``None`` if unknown)
        """
        # b2_copy_part doesn't need SSE-B2. Large file encryption is decided on b2_start_large_file.
        if (
            destination_encryption is not None
            and destination_encryption.mode == EncryptionMode.SSE_B2
        ):
            destination_encryption = None

        # Check if this part was uploaded before
        if finished_parts is not None and part_number in finished_parts:
            # Report this part finished
            part = finished_parts[part_number]
            large_file_upload_state.update_part_bytes(part.content_length)

            # Return SHA1 hash
            return {'contentSha1': part.content_sha1}

        # if another part has already had an error there's no point in
        # uploading this part
        if large_file_upload_state.has_error():
            raise AlreadyFailed(large_file_upload_state.get_error_message())

        response = self.services.session.copy_part(
            part_copy_source.file_id,
            large_file_id,
            part_number,
            bytes_range=part_copy_source.get_bytes_range(),
            destination_server_side_encryption=destination_encryption,
            source_server_side_encryption=source_encryption,
        )
        large_file_upload_state.update_part_bytes(response['contentLength'])
        return response

    def _copy_small_file(
        self,
        copy_source,
        file_name,
        content_type,
        file_info,
        destination_bucket_id,
        progress_listener: AbstractProgressListener,
        destination_encryption: EncryptionSetting | None,
        source_encryption: EncryptionSetting | None,
        legal_hold: LegalHold | None = None,
        file_retention: FileRetentionSetting | None = None,
    ):
        progress_listener.set_total_bytes(copy_source.get_content_length() or 0)

        bytes_range = copy_source.get_bytes_range()

        if content_type is None:
            if file_info is not None:
                raise CopyArgumentsMismatch('File info can be set only when content type is set')
            metadata_directive = MetadataDirectiveMode.COPY
        else:
            if file_info is None:
                raise CopyArgumentsMismatch(
                    'File info can be not set only when content type is not set'
                )
            metadata_directive = MetadataDirectiveMode.REPLACE
        metadata_directive, file_info, content_type = self.establish_sse_c_file_metadata(
            metadata_directive=metadata_directive,
            destination_file_info=file_info,
            destination_content_type=content_type,
            destination_server_side_encryption=destination_encryption,
            source_server_side_encryption=source_encryption,
            source_file_info=copy_source.source_file_info,
            source_content_type=copy_source.source_content_type,
        )
        response = self.services.session.copy_file(
            copy_source.file_id,
            file_name,
            bytes_range=bytes_range,
            metadata_directive=metadata_directive,
            content_type=content_type,
            file_info=file_info,
            destination_bucket_id=destination_bucket_id,
            destination_server_side_encryption=destination_encryption,
            source_server_side_encryption=source_encryption,
            legal_hold=legal_hold,
            file_retention=file_retention,
        )
        file_version = self.services.api.file_version_factory.from_api_response(response)
        progress_listener.bytes_completed(file_version.size)

        return file_version

    @classmethod
    def establish_sse_c_file_metadata(
        cls,
        metadata_directive: MetadataDirectiveMode,
        destination_file_info: dict | None,
        destination_content_type: str | None,
        destination_server_side_encryption: EncryptionSetting | None,
        source_server_side_encryption: EncryptionSetting | None,
        source_file_info: dict | None,
        source_content_type: str | None,
    ):
        assert metadata_directive in (MetadataDirectiveMode.REPLACE, MetadataDirectiveMode.COPY)

        if metadata_directive == MetadataDirectiveMode.REPLACE:
            if destination_server_side_encryption:
                destination_file_info = destination_server_side_encryption.add_key_id_to_file_info(
                    destination_file_info
                )
            return metadata_directive, destination_file_info, destination_content_type

        source_key_id = None
        destination_key_id = None

        if (
            destination_server_side_encryption is not None
            and destination_server_side_encryption.key is not None
            and destination_server_side_encryption.key.key_id is not None
        ):
            destination_key_id = destination_server_side_encryption.key.key_id

        if (
            source_server_side_encryption is not None
            and source_server_side_encryption.key is not None
            and source_server_side_encryption.key.key_id is not None
        ):
            source_key_id = source_server_side_encryption.key.key_id

        if source_key_id == destination_key_id:
            return metadata_directive, destination_file_info, destination_content_type

        if source_file_info is None or source_content_type is None:
            raise SSECKeyIdMismatchInCopy(
                f'attempting to copy file using {MetadataDirectiveMode.COPY} without providing source_file_info '
                f'and source_content_type for differing sse_c_key_ids: source="{source_key_id}", '
                f'destination="{destination_key_id}"'
            )

        destination_file_info = source_file_info.copy()
        destination_file_info.pop(SSE_C_KEY_ID_FILE_INFO_KEY_NAME, None)
        if destination_server_side_encryption:
            destination_file_info = destination_server_side_encryption.add_key_id_to_file_info(
                destination_file_info
            )
        destination_content_type = source_content_type

        return MetadataDirectiveMode.REPLACE, destination_file_info, destination_content_type



######################################################################
#
# File: b2sdk/_internal/requests/included_source_meta.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from b2sdk._internal.included_sources import IncludedSourceMeta, add_included_source

included_source_meta = IncludedSourceMeta(
    'requests',
    'Included in a revised form',
    {
        'NOTICE': """Requests
Copyright 2019 Kenneth Reitz

Copyright 2021 Backblaze Inc.
Changes made to the original source:
requests.models.Response.iter_content has been overridden to pass `decode_content=False` argument to `self.raw.stream`
in order to NOT decompress data based on Content-Encoding header"""
    },
)
add_included_source(included_source_meta)



######################################################################
#
# File: b2sdk/_internal/transfer/transfer_manager.py
#
# Copyright 2022 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations


class TransferManager:
    """
    Base class for manager classes (copy, upload, download)
    """

    def __init__(self, services, **kwargs):
        self.services = services
        super().__init__(**kwargs)



######################################################################
#
# File: b2sdk/_internal/transfer/emerge/planner/upload_subpart.py
#
# Copyright 2020 Backblaze Inc. All Rights Reserved.
#
# License https://www.backblaze.com/using_b2_code.html
#
######################################################################
from __future__ import annotations

import io
from abc import ABCMeta, abstractmethod
from functools import partial

from b2sdk._internal.stream.chained import StreamOpener
from b2sdk._internal.stream.range import wrap_with_range
from b2sdk._internal.utils import hex_sha1_of_unlimited_stream


class BaseUploadSubpart(metaclass=ABCMeta):
    def __init__(self, outbound_source, relative_offset, length):
        self.outbound_source = outbound_source
        self.relative_offset = relative_offset
        self.length = length

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} outbound_source={repr(self.outbound_source)} relative_offset={self.relative_offset} '
            f'length={self.length}>'
        )

    @abstractmethod
    def get_subpart_id(self):
        pass

    @abstractmethod
    def get_stream_opener(self, emerge_execution=None):
        pass

    def is_hashable(self):
        return False


class RemoteSourceUploadSubpart(BaseUploadSubpart):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._download_buffer_cache = None

    def get_subpart_id(self):
        return (self.outbound_source.file_id, self.relative_offset, self.length)

    def get_stream_opener(self, emerge_execution=None):
        if emerge_execution is None:
            raise RuntimeError('Cannot open remote source without emerge execution instance.')
        return CachedBytesStreamOpener(partial(self._download, emerge_execution))

    def _download(self, emerge_execution):
        url = emerge_execution.services.session.get_download_url_by_id(self.outbound_source.file_id)
        absolute_offset = self.outbound_source.offset + self.relative_offset
        range_ = (absolute_offset, absolute_offset + self.length - 1)
        with io.BytesIO() as bytes_io:
            downloaded_file = emerge_execution.services.download_manager.download_file_from_url(
                url, range_=range_, encryption=self.outbound_source.encryption
            )
            downloaded_file.save(bytes_io)
            return bytes_io.getvalue()


class LocalSourceUploadSubpart(BaseUploadSubpart):
    def get_subpart_id(self):
        with self._get_stream() as stream:
            sha1, _ = hex_sha1_of_unlimited_stream(stream)
            return sha1

    def get_stream_opener(self, emerge_execution=None):
        return self._get_stream

    def _get_stream(self):
        fp = self.outbound_source.open()
        return wrap_with_range(
            fp, self.outbound_source.get_content_length(), self.relative_offset, self.length
        )

    def is_hashable(self):
        return True


class CachedBytesStreamOpener(StreamOpener):
    def __init__(self, bytes_data_callback):
        self.bytes_data_callback = bytes_data_callback
        self._bytes_data_cache = None

    def __call__(self):
        if self._bytes_data_cache is None:
            self._bytes_data_cache = self.bytes_data_callback()
        return io.BytesIO(self._bytes_data_cache)

    def cleanup(self):
        self._bytes_data_cache = None
